#include "model.h"

#include <QSize>

ModelWindow::ModelWindow(QWindow *parent)
    : QWindow(parent)
{
    setSurfaceType(QWindow::OpenGLSurface);

}

/**
ModelWindow::~ModelWindow()
{
    delete m_device;
}
**/


void ModelWindow::initialize( )
{
    m_program = new QOpenGLShaderProgram(this);

    m_program->addShaderFromSourceFile(QOpenGLShader::Vertex, "vertexShaderSource.vertexshader" );
    m_program->addShaderFromSourceFile(QOpenGLShader::Fragment, "texture.fragmentshader" );  
    m_program->link();

    m_posAttr = m_program->attributeLocation("posAttr");
    Q_ASSERT(m_posAttr != -1);

    m_matrixUniform = m_program->uniformLocation("matrix");
    Q_ASSERT(m_matrixUniform != -1);

    m_texCoord = m_program->attributeLocation("texCoord");
    Q_ASSERT(m_texCoord != -1);

	m_program->bind();
	m_program->setUniformValue("texture", 0);

	inicializa_matrices();

	// Imagen de la textura
	if ( img_fondo.load( "modelo.png", "PNG" ) == false ) {
		fprintf( stderr, "ERROR: no se pudo cargar la imagen de fondo\n" );
	}
	texture = new QOpenGLTexture( img_fondo.mirrored() );

	/**
	// glGenTextures(1, &textureID);
	glBindTexture(GL_TEXTURE_2D, m_texture );
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, img_fondo.width(), img_fondo.height(), 0, GL_RGB, GL_UNSIGNED_BYTE, img_fondo.bits() );
	// Se puede borrar la imagen de textura aquí, OpenGL ya tiene una copia
	// glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	// glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER );
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER );
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	// glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
	// ... which requires mipmaps. Generate them automatically.
	// glGenerateMipmap(GL_TEXTURE_2D);
	**/
}


//! [5]
void ModelWindow::render( )
{
	// glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    const qreal retinaScale = devicePixelRatio();
    glViewport(0, 0, width() * retinaScale, height() * retinaScale);

    
	// Se dibuja la imagen de fondo
    m_program->bind();

  
    QMatrix4x4 matrix, M;
	// QMatrix4x4 view;
	// view.lookAt( QVector3D( 255.5, 255.5, 255.5 ), QVector3D( 255.5, 255.50, 0.0 ), QVector3D( 0.0, 1.0, 0.0)  );

    matrix.ortho(0.0, 511.0, 0.0, 511.0, -1, 800);
	// M = matrix*view;
    
    m_program->setUniformValue(m_matrixUniform, matrix);
  
    static const GLfloat vertices_back[] = {
            0.0f,     0.0f,
          511.0f,     0.0f,
          511.0f,   511.0f,
            0.0f,     0.0f,
          511.0f,   511.0f,
            0.0f,   511.0f
    };
    static const GLfloat uvs_back[] = {
         0.0f,  0.0f,
         1.0f,  0.0f,
         1.0f,  1.0f,
         0.0f,  0.0f,
         1.0f,  1.0f,
         0.0f,  1.0f
    };

    // m_program->setUniformValue(m_colorUniform, QColor( 210, 210, 210, 0) );
    glEnableVertexAttribArray(m_posAttr);
	// fprintf( stderr, "%d\n", m_posAttr );
    glVertexAttribPointer(m_posAttr, 2, GL_FLOAT, GL_FALSE, 0, vertices_back);

	glEnableVertexAttribArray(m_texCoord);
	// fprintf( stderr, "%d\n", m_texCoord );
    glVertexAttribPointer( m_texCoord, 2, GL_FLOAT, GL_FALSE, 0, uvs_back );

	texture->bind();
    glDrawArrays(GL_TRIANGLES, 0, 6);

    glDisableVertexAttribArray(m_posAttr);
    glDisableVertexAttribArray(m_texCoord);

	/** El objeto **/

    M = matrix * K * RT;
    
    m_program->setUniformValue(m_matrixUniform, M);

    static const GLfloat vertices[] = {
         -10.0f,  -10.0f,
          10.0f,  -10.0f,
          10.0f,   10.0f,
         -10.0f,  -10.0f,
          10.0f,   10.0f,
         -10.0f,   10.0f
    };

    // m_program->setUniformValue(m_colorUniform, QColor( 255, 255, 0, 0) );

    glVertexAttribPointer(m_posAttr, 2, GL_FLOAT, GL_FALSE, 0, vertices);
    glEnableVertexAttribArray(m_posAttr);

	glEnableVertexAttribArray(m_texCoord);
    glVertexAttribPointer( m_texCoord, 2, GL_FLOAT, GL_FALSE, 0, uvs_back );

	texture->bind();
    glDrawArrays(GL_TRIANGLES, 0, 6);

    glDisableVertexAttribArray(m_posAttr);
    glDisableVertexAttribArray(m_texCoord);

    m_program->release();

    ++m_frame;
}
//! [5]

void ModelWindow::inicializa_matrices()
{
    MK[0]  = f;    MK[1]  = 0.0; MK[2]  = -255.0; MK[3]  = 0.0;  
    MK[4]  = 0.0;  MK[5]  = f;   MK[6]  = -255.0; MK[7]  = 0.0;  
    MK[8]  = 0.0;  MK[9]  = 0.0; MK[10] = 101.0;  MK[11] = 100.0;  
    MK[12] = 0.0;  MK[13] = 0.0; MK[14] = -1.0;   MK[15] = 0.0;  

    K = QMatrix4x4 (MK);
    
    MRT[0 ] = 0.9736;  MRT[1 ] = 0.2282;  MRT[2 ] = -0.0002;  MRT[3 ] =  0.0382 ;    
    MRT[4 ] = -0.0895; MRT[5 ] = 0.3809;  MRT[6 ] =  0.9203;  MRT[7 ] = -0.025; 
    MRT[8 ] = 0.21;    MRT[9 ] = -0.896;  MRT[10] = 0.3913;   MRT[11] = -83.6709;
    MRT[12 ] = 0.0;    MRT[13 ] = 0.0;    MRT[14] = 0.0;      MRT[15] = 1.0;
    RT = QMatrix4x4 (MRT);

	/**
	const float * p = RT.constData();
	for( int i=0; i<16; i++ )
		printf("%g ", p[i] );
	printf("\n" );
	**/
}

void ModelWindow::keyPressEvent( QKeyEvent *e )
{
	switch ( e->key() ) {
		case Qt::Key_Escape:
		case Qt::Key_Q:
			QCoreApplication::quit();
			break;
	}
}

bool ModelWindow::event(QEvent *event)
{   
    switch (event->type()) {
    case QEvent::UpdateRequest:
        renderNow();
        return true;
    default:
        return QWindow::event(event);
    }
}

void ModelWindow::exposeEvent(QExposeEvent *event)
{   
    Q_UNUSED(event);
    
    if (isExposed())
        renderNow();
}

void ModelWindow::renderNow()
{
    if (!isExposed())
        return;

    bool needsInitialize = false;

    if (!m_context) {
        m_context = new QOpenGLContext(this);
        m_context->setFormat(requestedFormat());
        m_context->create();

        needsInitialize = true;
    }

    m_context->makeCurrent(this);

    if (needsInitialize) {
        initializeOpenGLFunctions();
		initialize( );
    }
    
    render();

    m_context->swapBuffers(this);
}
